# Copyright 2021 The alphaclops Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import sympy
import alphaclops
from alphaclops.ops.gateset_test import CustomX, CustomXPowGate


class UnitaryGate(alphaclops.Gate):
    def __init__(self, num_qubits: int) -> None:
        self._num_qubits = num_qubits

    def _has_unitary_(self) -> bool:
        return True

    def _num_qubits_(self) -> int:
        return self._num_qubits


def test_any_unitary_gate_family():
    with pytest.raises(ValueError, match='must be a positive integer'):
        _ = alphaclops.AnyUnitaryGateFamily(0)
    eq = alphaclops.testing.EqualsTester()
    eq.add_equality_group(alphaclops.AnyUnitaryGateFamily())
    for num_qubits in range(1, 6, 2):
        q = alphaclops.LineQubit.range(num_qubits)
        gate = UnitaryGate(num_qubits)
        eq.add_equality_group(alphaclops.AnyUnitaryGateFamily(num_qubits))
        for init_num_qubits in [None, num_qubits]:
            gate_family = alphaclops.AnyUnitaryGateFamily(init_num_qubits)
            alphaclops.testing.assert_equivalent_repr(gate_family)
            assert gate in gate_family
            assert gate(*q) in gate_family
            if init_num_qubits:
                assert f'{init_num_qubits}' in gate_family.name
                assert f'{init_num_qubits}' in gate_family.description
                assert UnitaryGate(num_qubits + 1) not in gate_family
            else:
                assert 'Any-Qubit' in gate_family.name
                assert 'any unitary' in gate_family.description

    assert alphaclops.testing.SingleQubitGate() not in alphaclops.AnyUnitaryGateFamily()


def test_any_integer_power_gate_family():
    with pytest.raises(ValueError, match='subclass of `alphaclops.EigenGate`'):
        alphaclops.AnyIntegerPowerGateFamily(gate=alphaclops.testing.SingleQubitGate)
    with pytest.raises(ValueError, match='subclass of `alphaclops.EigenGate`'):
        alphaclops.AnyIntegerPowerGateFamily(gate=CustomXPowGate())
    eq = alphaclops.testing.EqualsTester()
    gate_family = alphaclops.AnyIntegerPowerGateFamily(CustomXPowGate)
    eq.add_equality_group(gate_family)
    eq.add_equality_group(alphaclops.AnyIntegerPowerGateFamily(alphaclops.EigenGate))
    alphaclops.testing.assert_equivalent_repr(gate_family)
    assert CustomX in gate_family
    assert CustomX**2 in gate_family
    assert CustomX**1.5 not in gate_family
    assert CustomX ** sympy.Symbol('theta') not in gate_family
    assert 'CustomXPowGate' in gate_family.name
    assert '`g.exponent` is an integer' in gate_family.description


@pytest.mark.parametrize('gate', [CustomX, alphaclops.ParallelGate(CustomX, 2), CustomXPowGate])
@pytest.mark.parametrize('name,description', [(None, None), ("Custom Name", "Custom Description")])
@pytest.mark.parametrize('max_parallel_allowed', [None, 3])
def test_parallel_gate_family(gate, name, description, max_parallel_allowed):
    gate_family = alphaclops.ParallelGateFamily(
        gate, name=name, description=description, max_parallel_allowed=max_parallel_allowed
    )
    alphaclops.testing.assert_equivalent_repr(gate_family)
    for gate_to_test in [CustomX, alphaclops.ParallelGate(CustomX, 2)]:
        assert gate_to_test in gate_family
        assert gate_to_test(*alphaclops.LineQubit.range(alphaclops.num_qubits(gate_to_test))) in gate_family

    if isinstance(gate, alphaclops.ParallelGate) and not max_parallel_allowed:
        assert gate_family._max_parallel_allowed == alphaclops.num_qubits(gate)
        assert alphaclops.ParallelGate(CustomX, 4) not in gate_family
    else:
        assert gate_family._max_parallel_allowed == max_parallel_allowed
        assert (alphaclops.ParallelGate(CustomX, 4) in gate_family) == (max_parallel_allowed is None)

    str_to_search = 'Custom' if name else 'Parallel'
    assert str_to_search in gate_family.name
    assert str_to_search in gate_family.description


def test_parallel_gate_family_eq():
    eq = alphaclops.testing.EqualsTester()
    for name, description in [(None, None), ("Custom Name", "Custom Description")]:
        eq.add_equality_group(
            alphaclops.ParallelGateFamily(
                CustomX, max_parallel_allowed=2, name=name, description=description
            ),
            alphaclops.ParallelGateFamily(
                alphaclops.ParallelGate(CustomX, 2), name=name, description=description
            ),
        )
        eq.add_equality_group(
            alphaclops.ParallelGateFamily(
                CustomXPowGate, max_parallel_allowed=2, name=name, description=description
            )
        )
        eq.add_equality_group(
            alphaclops.ParallelGateFamily(
                CustomX, max_parallel_allowed=5, name=name, description=description
            ),
            alphaclops.ParallelGateFamily(
                alphaclops.ParallelGate(CustomX, 10),
                max_parallel_allowed=5,
                name=name,
                description=description,
            ),
        )
