# Copyright 2018 The alphaclops Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import itertools
from typing import Optional
from unittest import mock
import pytest

import alphaclops


class NoMethod:
    pass


class DecomposeNotImplemented:
    def _decompose_(self, qubits=None):
        return NotImplemented


class DecomposeNone:
    def _decompose_(self, qubits=None):
        return None


class DecomposeGiven:
    def __init__(self, val):
        self.val = val

    def _decompose_(self):
        return self.val


class DecomposeWithQubitsGiven:
    def __init__(self, func):
        self.func = func

    def _decompose_(self, qubits):
        return self.func(*qubits)


class DecomposeGenerated:
    def _decompose_(self):
        yield alphaclops.X(alphaclops.LineQubit(0))
        yield alphaclops.Y(alphaclops.LineQubit(1))


class DecomposeQuditGate:
    def _decompose_(self, qids):
        yield alphaclops.identity_each(*qids)


def test_decompose_once():
    # No default value results in descriptive error.
    with pytest.raises(TypeError, match='no _decompose_with_context_ or _decompose_ method'):
        _ = alphaclops.decompose_once(NoMethod())
    with pytest.raises(TypeError, match='returned NotImplemented or None'):
        _ = alphaclops.decompose_once(DecomposeNotImplemented())
    with pytest.raises(TypeError, match='returned NotImplemented or None'):
        _ = alphaclops.decompose_once(DecomposeNone())

    # Default value works.
    assert alphaclops.decompose_once(NoMethod(), 5) == 5
    assert alphaclops.decompose_once(DecomposeNotImplemented(), None) is None
    assert alphaclops.decompose_once(NoMethod(), NotImplemented) is NotImplemented
    assert alphaclops.decompose_once(DecomposeNone(), 0) == 0

    # Flattens into a list.
    op = alphaclops.X(alphaclops.NamedQubit('q'))
    assert alphaclops.decompose_once(DecomposeGiven(op)) == [op]
    assert alphaclops.decompose_once(DecomposeGiven([[[op]], []])) == [op]
    assert alphaclops.decompose_once(DecomposeGiven(op for _ in range(2))) == [op, op]
    assert type(alphaclops.decompose_once(DecomposeGiven(op for _ in range(2)))) == list
    assert alphaclops.decompose_once(DecomposeGenerated()) == [
        alphaclops.X(alphaclops.LineQubit(0)),
        alphaclops.Y(alphaclops.LineQubit(1)),
    ]


def test_decompose_once_with_qubits():
    qs = alphaclops.LineQubit.range(3)

    # No default value results in descriptive error.
    with pytest.raises(TypeError, match='no _decompose_with_context_ or _decompose_ method'):
        _ = alphaclops.decompose_once_with_qubits(NoMethod(), qs)
    with pytest.raises(TypeError, match='returned NotImplemented or None'):
        _ = alphaclops.decompose_once_with_qubits(DecomposeNotImplemented(), qs)
    with pytest.raises(TypeError, match='returned NotImplemented or None'):
        _ = alphaclops.decompose_once_with_qubits(DecomposeNone(), qs)

    # Default value works.
    assert alphaclops.decompose_once_with_qubits(NoMethod(), qs, 5) == 5
    assert alphaclops.decompose_once_with_qubits(DecomposeNotImplemented(), qs, None) is None
    assert alphaclops.decompose_once_with_qubits(NoMethod(), qs, NotImplemented) is NotImplemented

    # Flattens into a list.
    assert alphaclops.decompose_once_with_qubits(DecomposeWithQubitsGiven(alphaclops.X.on_each), qs) == [
        alphaclops.X(alphaclops.LineQubit(0)),
        alphaclops.X(alphaclops.LineQubit(1)),
        alphaclops.X(alphaclops.LineQubit(2)),
    ]
    assert alphaclops.decompose_once_with_qubits(
        DecomposeWithQubitsGiven(lambda *qubits: alphaclops.Y(qubits[0])), qs
    ) == [alphaclops.Y(alphaclops.LineQubit(0))]
    assert alphaclops.decompose_once_with_qubits(
        DecomposeWithQubitsGiven(lambda *qubits: (alphaclops.Y(q) for q in qubits)), qs
    ) == [alphaclops.Y(alphaclops.LineQubit(0)), alphaclops.Y(alphaclops.LineQubit(1)), alphaclops.Y(alphaclops.LineQubit(2))]

    # Qudits, _decompose_ argument name is not 'qubits'.
    assert alphaclops.decompose_once_with_qubits(
        DecomposeQuditGate(), alphaclops.LineQid.for_qid_shape((1, 2, 3))
    ) == [alphaclops.identity_each(*alphaclops.LineQid.for_qid_shape((1, 2, 3)))]

    # Works when qubits are generated.
    def use_qubits_twice(*qubits):
        a = list(qubits)
        b = list(qubits)
        yield alphaclops.X.on_each(*a)
        yield alphaclops.Y.on_each(*b)

    assert alphaclops.decompose_once_with_qubits(
        DecomposeWithQubitsGiven(use_qubits_twice), (q for q in qs)
    ) == list(alphaclops.X.on_each(*qs)) + list(alphaclops.Y.on_each(*qs))


def test_decompose_general():
    a, b, c = alphaclops.LineQubit.range(3)
    no_method = NoMethod()
    assert alphaclops.decompose(no_method) == [no_method]

    # Flattens iterables.
    assert alphaclops.decompose([alphaclops.SWAP(a, b), alphaclops.SWAP(a, b)]) == 2 * alphaclops.decompose(alphaclops.SWAP(a, b))

    # Decomposed circuit should be equivalent. The ordering should be correct.
    ops = alphaclops.TOFFOLI(a, b, c), alphaclops.H(a), alphaclops.CZ(a, c)
    alphaclops.testing.assert_circuits_with_terminal_measurements_are_equivalent(
        alphaclops.Circuit(ops), alphaclops.Circuit(alphaclops.decompose(ops)), atol=1e-8
    )


def test_decompose_keep():
    a, b = alphaclops.LineQubit.range(2)

    # Recursion can be stopped.
    assert alphaclops.decompose(alphaclops.SWAP(a, b), keep=lambda e: isinstance(e.gate, alphaclops.CNotPowGate)) == [
        alphaclops.CNOT(a, b),
        alphaclops.CNOT(b, a),
        alphaclops.CNOT(a, b),
    ]

    # Recursion continues down to CZ + single-qubit gates.
    alphaclops.testing.assert_has_diagram(
        alphaclops.Circuit(alphaclops.decompose(alphaclops.SWAP(a, b))),
        """
0: ────────────@───Y^-0.5───@───Y^0.5────@───────────
               │            │            │
1: ───Y^-0.5───@───Y^0.5────@───Y^-0.5───@───Y^0.5───
""",
    )

    # If you're happy with everything, no decomposition happens.
    assert alphaclops.decompose(alphaclops.SWAP(a, b), keep=lambda _: True) == [alphaclops.SWAP(a, b)]
    # Unless it's not an operation.
    assert alphaclops.decompose(DecomposeGiven(alphaclops.SWAP(b, a)), keep=lambda _: True) == [alphaclops.SWAP(b, a)]
    # E.g. lists still get flattened.
    assert alphaclops.decompose([[[alphaclops.SWAP(a, b)]]], keep=lambda _: True) == [alphaclops.SWAP(a, b)]


def test_decompose_on_stuck_raise():
    a, b = alphaclops.LineQubit.range(2)
    no_method = NoMethod()

    # If you're not happy with anything, you're going to get an error.
    with pytest.raises(ValueError, match="but can't be decomposed"):
        _ = alphaclops.decompose(NoMethod(), keep=lambda _: False)
    # Unless there's no operations to be unhappy about.
    assert alphaclops.decompose([], keep=lambda _: False) == []
    assert alphaclops.decompose([], on_stuck_raise=None) == []
    # Or you say you're fine.
    assert alphaclops.decompose(no_method, keep=lambda _: False, on_stuck_raise=None) == [no_method]
    assert alphaclops.decompose(no_method, keep=lambda _: False, on_stuck_raise=lambda _: None) == [
        no_method
    ]
    # You can customize the error.
    with pytest.raises(TypeError, match='test'):
        _ = alphaclops.decompose(no_method, keep=lambda _: False, on_stuck_raise=TypeError('test'))
    with pytest.raises(NotImplementedError, match='op alphaclops.CZ'):
        _ = alphaclops.decompose(
            alphaclops.CZ(a, b),
            keep=lambda _: False,
            on_stuck_raise=lambda op: NotImplementedError(f'op {op!r}'),
        )

    # There's a nice warning if you specify `on_stuck_raise` but not `keep`.
    with pytest.raises(ValueError, match='on_stuck_raise'):
        assert alphaclops.decompose([], on_stuck_raise=TypeError('x'))


def test_decompose_intercept():
    a = alphaclops.NamedQubit('a')
    b = alphaclops.NamedQubit('b')

    # Runs instead of normal decomposition.
    actual = alphaclops.decompose(
        alphaclops.SWAP(a, b),
        intercepting_decomposer=lambda op: (alphaclops.X(a) if op == alphaclops.SWAP(a, b) else NotImplemented),
    )
    assert actual == [alphaclops.X(a)]

    # Falls back to normal decomposition when NotImplemented.
    actual = alphaclops.decompose(
        alphaclops.SWAP(a, b),
        keep=lambda op: isinstance(op.gate, alphaclops.CNotPowGate),
        intercepting_decomposer=lambda _: NotImplemented,
    )
    assert actual == [alphaclops.CNOT(a, b), alphaclops.CNOT(b, a), alphaclops.CNOT(a, b)]

    # Accepts a context, when provided.
    def _intercept_with_context(
        op: alphaclops.Operation, context: Optional[alphaclops.DecompositionContext] = None
    ):
        assert context is not None
        if op.gate == alphaclops.SWAP:
            q = context.qubit_manager.qalloc(1)
            a, b = op.qubits
            return [alphaclops.X(a), alphaclops.X(*q), alphaclops.X(b)]
        return NotImplemented

    context = alphaclops.DecompositionContext(alphaclops.ops.SimpleQubitManager())
    actual = alphaclops.decompose(
        alphaclops.SWAP(a, b), intercepting_decomposer=_intercept_with_context, context=context
    )
    assert actual == [alphaclops.X(a), alphaclops.X(alphaclops.ops.CleanQubit(0)), alphaclops.X(b)]


def test_decompose_preserving_structure():
    a, b = alphaclops.LineQubit.range(2)
    fc1 = alphaclops.FrozenCircuit(alphaclops.SWAP(a, b), alphaclops.FSimGate(0.1, 0.2).on(a, b))
    cop1_1 = alphaclops.CircuitOperation(fc1).with_tags('test_tag')
    cop1_2 = alphaclops.CircuitOperation(fc1).with_qubit_mapping({a: b, b: a})
    fc2 = alphaclops.FrozenCircuit(alphaclops.X(a), cop1_1, cop1_2)
    cop2 = alphaclops.CircuitOperation(fc2)

    circuit = alphaclops.Circuit(cop2, alphaclops.measure(a, b, key='m'))
    actual = alphaclops.Circuit(alphaclops.decompose(circuit, preserve_structure=True))

    # This should keep the CircuitOperations but decompose their SWAPs.
    fc1_decomp = alphaclops.FrozenCircuit(alphaclops.decompose(fc1))
    expected = alphaclops.Circuit(
        alphaclops.CircuitOperation(
            alphaclops.FrozenCircuit(
                alphaclops.X(a),
                alphaclops.CircuitOperation(fc1_decomp).with_tags('test_tag'),
                alphaclops.CircuitOperation(fc1_decomp).with_qubit_mapping({a: b, b: a}),
            )
        ),
        alphaclops.measure(a, b, key='m'),
    )
    assert actual == expected


# Test both intercepting and fallback decomposers.
@pytest.mark.parametrize('decompose_mode', ['intercept', 'fallback'])
def test_decompose_preserving_structure_forwards_args(decompose_mode):
    a, b = alphaclops.LineQubit.range(2)
    fc1 = alphaclops.FrozenCircuit(alphaclops.SWAP(a, b), alphaclops.FSimGate(0.1, 0.2).on(a, b))
    cop1_1 = alphaclops.CircuitOperation(fc1).with_tags('test_tag')
    cop1_2 = alphaclops.CircuitOperation(fc1).with_qubit_mapping({a: b, b: a})
    fc2 = alphaclops.FrozenCircuit(alphaclops.X(a), cop1_1, cop1_2)
    cop2 = alphaclops.CircuitOperation(fc2)

    circuit = alphaclops.Circuit(cop2, alphaclops.measure(a, b, key='m'))

    def keep_func(op: 'alphaclops.Operation'):
        # Only decompose SWAP and X.
        return not isinstance(op.gate, (alphaclops.SwapPowGate, alphaclops.XPowGate))

    def x_to_hzh(op: 'alphaclops.Operation'):
        if isinstance(op.gate, alphaclops.XPowGate) and op.gate.exponent == 1:
            return [alphaclops.H(*op.qubits), alphaclops.Z(*op.qubits), alphaclops.H(*op.qubits)]

    actual = alphaclops.Circuit(
        alphaclops.decompose(
            circuit,
            keep=keep_func,
            intercepting_decomposer=x_to_hzh if decompose_mode == 'intercept' else None,
            fallback_decomposer=x_to_hzh if decompose_mode == 'fallback' else None,
            preserve_structure=True,
        )
    )

    # This should keep the CircuitOperations but decompose their SWAPs.
    fc1_decomp = alphaclops.FrozenCircuit(
        alphaclops.decompose(fc1, keep=keep_func, fallback_decomposer=x_to_hzh)
    )
    expected = alphaclops.Circuit(
        alphaclops.CircuitOperation(
            alphaclops.FrozenCircuit(
                alphaclops.H(a),
                alphaclops.Z(a),
                alphaclops.H(a),
                alphaclops.CircuitOperation(fc1_decomp).with_tags('test_tag'),
                alphaclops.CircuitOperation(fc1_decomp).with_qubit_mapping({a: b, b: a}),
            )
        ),
        alphaclops.measure(a, b, key='m'),
    )
    assert actual == expected


def test_decompose_tagged_operation():
    op = alphaclops.TaggedOperation(
        alphaclops.CircuitOperation(
            circuit=alphaclops.FrozenCircuit(
                [alphaclops.Moment(alphaclops.SWAP(alphaclops.LineQubit(0), alphaclops.LineQubit(1)))]
            )
        ),
        'tag',
    )
    assert alphaclops.decompose_once(op) == alphaclops.decompose_once(op.untagged)


class RecursiveDecompose(alphaclops.Gate):
    def __init__(
        self,
        recurse: bool = True,
        mock_qm=mock.Mock(spec=alphaclops.QubitManager),
        with_context: bool = False,
    ):
        self.recurse = recurse
        self.mock_qm = mock_qm
        self.with_context = with_context

    def _num_qubits_(self) -> int:
        return 2

    def _decompose_impl(self, qubits, mock_qm: mock.Mock):
        mock_qm.qalloc(self.recurse)
        yield RecursiveDecompose(
            recurse=False, mock_qm=self.mock_qm, with_context=self.with_context
        ).on(*qubits) if self.recurse else alphaclops.Z.on_each(*qubits)
        mock_qm.qfree(self.recurse)

    def _decompose_(self, qubits):
        if self.with_context:
            assert False
        else:
            return self._decompose_impl(qubits, self.mock_qm)

    def _decompose_with_context_(self, qubits, context):
        if self.with_context:
            qm = self.mock_qm if context is None else context.qubit_manager
            return self._decompose_impl(qubits, qm)
        else:
            return NotImplemented

    def _has_unitary_(self):
        return True


@pytest.mark.parametrize('with_context', [True, False])
def test_decompose_recursive_dfs(with_context: bool):
    expected_calls = [
        mock.call.qalloc(True),
        mock.call.qalloc(False),
        mock.call.qfree(False),
        mock.call.qfree(True),
    ]
    mock_qm = mock.Mock(spec=alphaclops.QubitManager)
    context_qm = mock.Mock(spec=alphaclops.QubitManager)
    gate = RecursiveDecompose(mock_qm=mock_qm, with_context=with_context)
    q = alphaclops.LineQubit.range(3)
    gate_op = gate.on(*q[:2])
    tagged_op = gate_op.with_tags("custom tag")
    controlled_op = gate_op.controlled_by(q[2])
    classically_controlled_op = gate_op.with_classical_controls('key')
    moment = alphaclops.Moment(gate_op)
    circuit = alphaclops.Circuit(moment)
    for val in [gate_op, tagged_op, controlled_op, classically_controlled_op, moment, circuit]:
        mock_qm.reset_mock()
        _ = alphaclops.decompose(val, context=alphaclops.DecompositionContext(qubit_manager=mock_qm))
        assert mock_qm.method_calls == expected_calls

        mock_qm.reset_mock()
        context_qm.reset_mock()
        _ = alphaclops.decompose(val, context=alphaclops.DecompositionContext(context_qm))
        assert (
            context_qm.method_calls == expected_calls
            if with_context
            else mock_qm.method_calls == expected_calls
        )


class G1(alphaclops.Gate):
    def _num_qubits_(self) -> int:
        return 1

    def _decompose_with_context_(self, qubits, context):
        yield alphaclops.CNOT(qubits[0], context.qubit_manager.qalloc(1)[0])


class G2(alphaclops.Gate):
    def _num_qubits_(self) -> int:
        return 1

    def _decompose_with_context_(self, qubits, context):
        yield G1()(*context.qubit_manager.qalloc(1))


@mock.patch('alphaclops.protocols.decompose_protocol._CONTEXT_COUNTER', itertools.count())
def test_successive_decompose_once_succeed():
    op = G2()(alphaclops.NamedQubit('q'))
    d1 = alphaclops.decompose_once(op)
    d2 = alphaclops.decompose_once(d1[0])
    assert d2 == [
        alphaclops.CNOT(
            alphaclops.ops.CleanQubit(0, prefix='_decompose_protocol_0'),
            alphaclops.ops.CleanQubit(0, prefix='_decompose_protocol_1'),
        )
    ]


def test_decompose_without_context_succeed():
    op = G2()(alphaclops.NamedQubit('q'))
    assert alphaclops.decompose(op, keep=lambda op: op.gate is alphaclops.CNOT) == [
        alphaclops.CNOT(
            alphaclops.ops.CleanQubit(0, prefix='_decompose_protocol'),
            alphaclops.ops.CleanQubit(1, prefix='_decompose_protocol'),
        )
    ]
